###########################################################
## This file contains the source code of the prototype
## version of the lgarch(p,q) source
##
## First created 29 June 2016
##
## 1 INITIATE, LOAD LIBRARIES
##
## 2 COMPILE C-FUNCTIONS
##
## 3 UNIVARIATE LOG-GARCH(P,Q) FUNCTIONS:
##   - lgarchmcRecursion         #recursion
##   - lgarchmcObjective         #objective function
##   - lgarchmc                 #estimate; return object from lgarch class
##
## 4 EXTRACTION FUNCTION (MOSTLY S3 METHODS) FOR LGARCHMC
##   - fitted.lgarchmc
##   - residuals.lgarchmc
##
###########################################################


##=========================================================
## 1 INITIATE, LOAD LIBRARIES
##=========================================================

##remove everything in workspace:
rm(list=ls())

##set working directory:
#setwd("C:/Users/sucarrat/Documents/R/gs/lgarch/prototype/")

##load libraries
library(zoo)
library(inline)

##=========================================================
## 2 C-FUNCTION
##=========================================================

mysig <- signature(iStart="integer", iEnd="integer",
  vY="numeric", vU="numeric", vX="numeric",
  phi="numeric", iARorder="numeric",
  theta="numeric", iMAorder="numeric", vMissing="numeric")
mycode <- "
  double phisum; phisum = 0;
  double thetasum; thetasum = 0;
  double vYhat;

  /*observation loop:*/
  for(int i=*iStart; i < *iEnd; i++){

      /*AR sum:*/
      if(iARorder > 0){
        phisum = 0;
        for(int j=0; j < *iARorder; j++){
          phisum = phisum + phi[j] * vY[i-1-j];
        }
      } /*end AR sum*/

      /*MA sum:*/
      if(iMAorder > 0){
        thetasum = 0;
        for(int j=0; j < *iMAorder; j++){
          thetasum = thetasum + theta[j] * vU[i-1-j];
        }
      } /*end MA sum*/

      /*Compute vYhat:*/
      vYhat = phisum + thetasum + vX[i];

      /*check for missing:*/
      if(vMissing[i] == 1){
        vY[i] = vYhat;
      }

      /*Sum the parts:*/
      vU[i] = vY[i] - vYhat;

  } /*end observation loop*/
"
ARMARECURSION <- cfunction(mysig, mycode, convention=".C") #compile

##clean-up:
rm(mysig, mycode)


##=========================================================
## 3 UNIVARIATE LOG-GARCH(P,Q) FUNCTIONS
##=========================================================

#################
lgarchmcRecursion <- function(pars, aux, verbose=FALSE)
{
  ##parameter vectors:
  phi <- 0
  if(aux$iARorder > 0){ phi <- pars[1:aux$iARorder] }
  theta <- 0
  if(aux$iMAorder > 0){ theta <- pars[c(aux$iARorder+1):c(aux$iARorder+aux$iMAorder)] }
  if(aux$xreg.k > 0){
    parsxreg <- pars[c(aux$iARorder+aux$iMAorder+1):length(pars)]
    aux$vX <- as.vector(aux$vX %*% cbind(parsxreg))
  }

  ##recursion:
  if(aux$c.code){
    tmp <- ARMARECURSION(as.integer(aux$iMAXorder),
      as.integer(aux$vYlength), as.numeric(aux$vY),
      as.numeric(aux$vU), as.numeric(aux$vX),
      as.numeric(phi), as.integer(aux$iARorder),
      as.numeric(theta), as.integer(aux$iMAorder),
      as.numeric(aux$vMissing))
    if(verbose){
      vU <- cbind(tmp$vU,tmp$vY,tmp$vX)
    }else{
      vU <- tmp$vU
    }
  }else{
    iStart <- aux$iMAXorder+1
    iEnd <- aux$vYlength
    phisum <- 0
    thetasum <- 0
    vY <- aux$vY
    vU <- aux$vU
    vX <- aux$vX
    vMissing <- aux$vMissing
    for(i in iStart:iEnd){
      if(aux$iARorder>0){
        phisum <- sum(phi*vY[c(i-1):c(i-aux$iARorder)])
      }
      if(aux$iMAorder>0){
        thetasum <- sum(theta*vU[c(i-1):c(i-aux$iMAorder)])
      }
      vYhat <- phisum + thetasum + vX[i]
      if(vMissing[i]==1){
        vY[i] <- vYhat
      }
      vU[i] <- vY[i] - vYhat
    } #end for loop

    if(verbose){
      vU <- cbind(vU,vY,vX)
    }

  } #end if(aux$c.code)

  ##return:
  if(aux$iMAXorder > 0){
    if(NCOL(vU)==1){ vU <- vU[-c(1:aux$iMAXorder)] }
    if(NCOL(vU) >1){ vU <- vU[-c(1:aux$iMAXorder),] }
  }
  return(vU)
} #end lgarchmcRecursion

#################
lgarchmcObjective <- function(pars, aux)
{
  #check parameters:
  if( any(is.na(pars)) || any(pars<=aux$lower) || any(pars>=aux$upper) ){
    checkConds <- FALSE
  }else{
    checkConds <- TRUE
  } #end if..else..

  if(checkConds){
    uadj <- lgarchmcRecursion(pars, aux) #recursion
    objectiveValue <- sum(uadj^2)
    if(is.na(objectiveValue) || abs(objectiveValue) == Inf){
      objectiveValue <- aux$objective.penalty
    }
  }else{
    objectiveValue <- aux$objective.penalty
  } #end if..else..

  return(objectiveValue)

} #end lgarchmcObjective

#################
lgarchmc <- function(y, arch=1, garch=1, asym=0,
  asym.threshold=0, asymext=0, asymext.threshold=0,
  xregdet=NULL, xreg=NULL, initial.values=NULL,
  lower=NULL, upper=NULL, control=list(), vcov=TRUE,
  method="ls", mean.correction=TRUE, objective.penalty=NULL,
  solve.tol=.Machine$double.eps, solve.qr.tol=1e-07,
  LAPACK=FALSE, c.code=TRUE)
{
  ##check/change arguments:

  ##name, zoo and xts specific:
#  y.name <- deparse(substitute(y))
  y <- as.zoo(cbind(y))
#  if( is.null(y.name)){ y.name <- colnames(y)[1] }
#  if( y.name[1] =="" ){ y.name <- "y" }
  y <- na.trim(y)
  y.index <- index(y)
  y.n <- NROW(y)
  t1 <- y.index[1]
  t2 <- y.index[y.n]
  y <- coredata(y)
  if(NCOL(y) > 1){
    stop("Dependent variable not 1-dimensional")
  }else{
    y <- y[,1]
  }
  lny2 <- log(y^2)

  ##create aux list:
  aux <- list()

  ##deterministic regressors: intercept and xregdet
  x <- cbind(rep(1,y.n))
  colnames(x)[1] <- "intercept"
  if(!is.null(xregdet)){
    xregdet <- as.zoo(cbind(xregdet))
    xregdet.names <- colnames(xregdet)
    if(is.null(xregdet.names)){
      xregdet.names <- paste("xregdet", 1:NCOL(xregdet), sep="")
    }
    if(any(xregdet.names == "")){
      missing.colnames <- which(xregdet.names == "")
      for(i in 1:length(missing.colnames)){
        xregdet.names[missing.colnames[i]] <- paste("xregdet", i, sep="")
      }
    }
    xregdet <- window(xregdet, start=t1, end=t2)
    xregdet <- cbind(coredata(xregdet))

    ##naming:
    colnames(xregdet) <- xregdet.names
    x <- cbind(x,xregdet)
  }
  xregdet.names <- colnames(x)
#  colnames(x) <- NULL #for faster computation?

  ##estimate lng (log of deterministic component):
  xtx <- crossprod(x)
  lng <- qr(xtx, solve.qr.tol, LAPACK=LAPACK)
  lng$xtxinv <- solve.qr(lng, tol=solve.qr.tol, LAPACK=LAPACK)
  lng$xtx <- xtx
  lng$xty <- crossprod(x, cbind(lny2))
  lng$coefficients <- as.vector(lng$xtxinv %*% lng$xty)
  lng$fitted <- as.vector(x %*% cbind(lng$coefficients))
  lng$residuals <- lny2 - lng$fitted
  aux$vY <- lng$residuals

  ##stochastic regressors: xreg
  if(!is.null(xreg)){
    xreg <- as.zoo(cbind(xreg))
    xreg.names <- colnames(xreg)
    if(is.null(xreg.names)){
      xreg.names <- paste("xreg", 1:NCOL(xreg), sep="")
    }
    if(any(xreg.names == "")){
      missing.colnames <- which(xreg.names == "")
      for(i in 1:length(missing.colnames)){
        xreg.names[i] <- paste("xreg", i, sep="")
      }
    }
    xreg <- window(xreg, start=t1, end=t2)
    xreg <- cbind(coredata(xreg))
    xregMeans <- colMeans(xreg)
    xregMeans <- t(matrix(rep(xregMeans), NCOL(xreg), NROW(xreg)))
    xreg <- xreg - xregMeans
  }

  ##arch, garch orders, indices:
  aux$iARorder <- max(arch)
  aux$iMAorder <- max(garch)
  aux$iMAXorder <- max(aux$iARorder, aux$iMAorder)
  if(aux$iARorder > 0){
    aux$ARindx <- 1:aux$iARorder
  }else{
    aux$ARindx <- 0
  }
  if(aux$iMAorder > 0){
    aux$MAindx <- c(max(aux$ARindx)+1):c(max(aux$ARindx)+aux$iMAorder)
  }else{
    aux$MAindx <- 0
  }
  if(is.null(xreg)){
    aux$xreg.k <- 0
  }else{
    aux$xreg.k <- NCOL(xreg)
    aux$xreg.indx <- c(max(aux$ARindx,aux$MAindx)+1):c(max(aux$ARindx,aux$MAindx)+aux$xreg.k)
  }

  ##initial values:
  if(aux$iARorder > 0){
    vARinitvals <- rep(0.9/aux$iARorder, aux$iARorder)
  }else{
    vARinitvals <- NULL #numeric(0) instead?
  }
  if(aux$iMAorder > 0){
    vMAinitvals <- rep(-0.8/aux$iMAorder, aux$iMAorder)
  }else{
    vMAinitvals <- NULL #numeric(0) instead?
  }
  if(is.null(xreg)){
    xregInitvals <- NULL #numeric(0) instead?
  }else{
    xregInitvals <- rep(0.01, NCOL(xreg))
  }
  initvals <- c(vARinitvals, vMAinitvals, xregInitvals)

  #to do: check for stability?

  #upper bounds:
  if(!is.null(vARinitvals)){
    vARupper <- rep(1-.Machine$double.eps, aux$iARorder)
  }else{
    vARupper <- NULL
  }
  if(!is.null(vARinitvals)){
    vMAupper <- rep(1-.Machine$double.eps, aux$iMAorder)
  }else{
    vMAupper <- NULL
  }
  if(is.null(xreg)){
    xregUpper <- NULL #numeric(0) instead?
  }else{
    xregUpper <- rep(Inf, NCOL(xreg))
  }
  aux$uppervals <- c(vARupper, vMAupper, xregUpper)

  #lower bounds:
  if(!is.null(vARinitvals)){
    vARlower <- rep(.Machine$double.eps-1, aux$iARorder)
  }else{
    vARlower <- NULL
  }
  if(!is.null(vARinitvals)){
    vMAlower <- rep(.Machine$double.eps-1, aux$iMAorder)
  }else{
    vMAlower <- NULL
  }
  if(is.null(xreg)){
    xregLower <- NULL #numeric(0) instead?
  }else{
    xregLower <- rep(-Inf, NCOL(xreg))
  }
  aux$lowervals <- c(vARlower, vMAlower, xregLower)

  ##vectors:
  if(aux$iMAXorder > 0){
    aux$vY <- c(rep(0, aux$iMAXorder), aux$vY)
  }
  aux$vYlength <- length(aux$vY)
  aux$vU <- rep(0, aux$vYlength)
  if(is.null(xreg)){
    aux$vX <- aux$vU
  }else{
    aux$vX <- rbind(matrix(0, aux$iMAXorder, NCOL(xreg)), xreg)
  }
  aux$vMissing <- aux$vU

  ##misc:
  aux$c.code <- c.code
  if(is.null(objective.penalty)){
    aux$objective.penalty <- lgarchmcObjective(initvals, aux)
  }else{
    aux$objective.penalty <- objective.penalty
  }

  ##estimate:
  if(method=="ls"){
    objectiveFun <- function(pars, x=aux){ lgarchmcObjective(pars,x) }
  }else{
    objectiveFun <- function(pars, x=aux){ -lgarchmcObjective(pars,x) }
  }
  est <- nlminb(initvals, objectiveFun, lower=aux$lower,
    upper=aux$upper, control=control)

  ##Elnz2 and sigma2u parameters:
  mUadj <- lgarchmcRecursion(est$par, aux, verbose=TRUE)
  colnames(mUadj) <- c("vU", "vY", "vX")
  Elnz2 <- -log(mean(exp(mUadj[,1])))
  namesArma <- NULL
  sigma2u <- var(mUadj[,1])
  par.lgarch <- Elnz2
  namesLgarch <- "Elnz2"
  parArmaMC <- est$par #store for vcov estimation (further below)
  par.arma <- est$par

  #xreg, garch, arch and intercept parameters:
  if(!is.null(xreg)){
    namesArma <- c(xreg.names, namesArma)
    namesLgarch <- c(xreg.names, namesLgarch)
    par.lgarch <- c(est$par[aux$xreg.indx], par.lgarch)
  }
  if(aux$iMAorder > 0){
    namesArma <- c(paste("ma",1:aux$iMAorder,sep=""), namesArma)
    namesLgarch <- c(paste("garch",1:aux$iMAorder,sep=""), namesLgarch)
    par.lgarch <- c(-est$par[aux$MAindx], par.lgarch)
  }
  if(aux$iARorder > 0){
    namesArma <- c(paste("ar",1:aux$iARorder,sep=""), namesArma)
    namesLgarch <- c(paste("arch",1:aux$iARorder,sep=""), namesLgarch)
    par.tmp <- c(est$par[aux$MAindx], rep(0, aux$iARorder-aux$iMAorder))
    par.lgarch <- c(est$par[aux$ARindx]+par.tmp, par.lgarch)
  }

  ##deterministic part:
  namesArma <- c(xregdet.names, namesArma)
  namesLgarch <- c(xregdet.names, namesLgarch)
  par.arma <- c(lng$coefficients, par.arma)
  par.lgarch <- c(lng$coefficients, par.lgarch)

  ##adjust intercept:

  ##completion of parameters:
  names(par.arma) <- namesArma
  names(par.lgarch) <- namesLgarch
  est$par <- par.lgarch
  est <- c(list(date=date(), par.arma=par.arma), est)

  ##fitted values, residuals, y:
  arma.fitted <- mUadj[,"vY"] - mUadj[,"vU"]
  sigma.fitted <- exp( (lng$fitted + arma.fitted - Elnz2)/2 )
  resids.std <- y/sigma.fitted
  arma.fitted <- zoo(arma.fitted, order.by=y.index)
  sigma.fitted <- zoo(sigma.fitted, order.by=y.index)
  resids.std <- zoo(resids.std, order.by=y.index)
  dependentvar <- zoo(y, order.by=y.index)
  est <- c(list(dependentvar=dependentvar,
    resids.std=resids.std, sd=sigma.fitted), est)

  ##out:
#  if(is.null(est$aux)){
#    est <- c(list(aux=aux),est)
#  }
  class(est) <- "lgarchmc"
  return(est)
}

##=========================================================
## 4 EXTRACTION FUNCTION (MOSTLY S3 METHODS) FOR LGARCHMC
##=========================================================

################### fitted values
#add verbose=FALSE argument in the future?
fitted.lgarchmc <- function(object, ...)
{
#for the future?:
#  if(verbose){
#    zhat <- object$resids.std
#    result <- cbind(sigma, logsigma2, zhat, mUhat[,1])
#    colnames(result) <- c("sd", "lnsd2", "zhat", "uhat")
#    result <- zoo(result, order.by=object$aux$y.index)
#  }else{
    result <- object$sd
#  }
  return(result)
} #end fitted.lgarchmc

################### fitted values
residuals.lgarchmc <- function(object, ...)
{
  return(object$resids.std)
} #end residuals.lgarchmc

################### fitted values
logLik.lgarchmc <- function(object, ...)
{
  sigma <- coredata(object$sd)
  result <- sum(dnorm(coredata(object$dependentvar),
      sd=sigma, log=TRUE))
  attr(result, "df") <- length(object$par)-1
  attr(result, "nobs") <- length(object$dependentvar)
  class(result) <- "logLik"
  return(result)
} #end logLik.lgarchmc
